/*
 *    Graph.java
 *    Copyright (C) 2000 Gilad Mishne
 *
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

package weka.core;

import java.util.Enumeration;
import java.util.Hashtable;
import java.util.Vector;
import java.util.Stack;

/**
 * Class for creating basic graphs and extracting
 * SCCs (strongly connected components) from them.
 * No support for deletion of elements from the graph.
 *
 * @author Gilad Mishne (gilad.mishne@intel.com)
 * @version $Revision: 1.0 $
 */
public class Graph  
{
    public class Vertex
    {
        private Object element;
        boolean isInGraph;
        boolean visited;
        Vertex succ,pred;
        Edge in,out; // pointer to cyclic list of in-Edges and out-Edges
    
        /** Takes no arguments.
         */
        Vertex() {
            isInGraph = true;   
            succ=this; // point back to itself
            pred=this;
            in = null;
            out = null;
        }
    
        /** Associates the given objct with the Vertex.
            @param obj The object.
         */
        Vertex(Object obj) {
            isInGraph = true;       
            element=obj;
            succ=this; // point back to itself
            pred=this;
            in = null;
            out = null;
        }
    
        /**
       @return The object associated with the Vertex
         */
        public Object element() {
            return element;
        }

        /** Associates the Vertex with the given object.
            @param obj The object.
    */
        public void element(Object obj) {
            element=obj;
        }
    }
    
    public class Edge 
    {
        boolean isInGraph;
        int visited;
    
        Edge succ, pred;
        Vertex source, target;
        Edge insucc, inpred; // reference to in-successor (pred.) in the cyclic list
        Edge outsucc,outpred;
    
        Edge() {
            isInGraph = true;
            source=target=null;
            succ=pred=this;
            insucc=inpred=this; // point back to itself
            outsucc=outpred=this; // point back to itself
        }
    
        /** Returns string representation of an Edge.
         */
        public String toString() {
            StringBuffer s = new StringBuffer();
            s.append(source.element.toString() + "->" + target.element.toString());
            return s.toString();
            
        }
        
    }
    
    class AdjVerticesEnumerator implements Enumeration {
    
    private Edge startEdge;
    private Edge currentEdge;
    
    AdjVerticesEnumerator(Vertex startVertex) {
        startEdge=startVertex.out;
        currentEdge = startEdge;
        }
    
    public boolean  hasMoreElements() {
            return currentEdge != null;
        }
    
    public Object   nextElement() {
        Object  anObject = currentEdge.target;
        currentEdge=currentEdge.outsucc;
        if(currentEdge==startEdge) // if we have seen all edges yet
        currentEdge=null;
        
        return  anObject;
    }
    }
    
    class AdjEdgesEnumerator implements Enumeration 
    {
    private Edge startEdge;
    private Edge currentEdge;
    AdjEdgesEnumerator(Vertex startVertex) {
        startEdge=startVertex.out;
        currentEdge = startEdge;
        }
    
    public boolean  hasMoreElements() {
            return currentEdge != null;
        }
    
    public Object   nextElement() {
        Object  anObject = currentEdge;
        currentEdge=currentEdge.outsucc;
        if(currentEdge==startEdge) // if we have seen all edges yet
        currentEdge=null;
        
        return  anObject;
    }
    }
    
    class InEdgesEnumerator implements Enumeration 
    {    
    private Edge startEdge;
    private Edge currentEdge;
    
    InEdgesEnumerator(Vertex startVertex) {
        startEdge=startVertex.in;
        currentEdge = startEdge;
        }
    
    public boolean  hasMoreElements() {
            return currentEdge != null;
        }
    
    public Object   nextElement() {
        Object  anObject = currentEdge;
        currentEdge=currentEdge.insucc;
        if(currentEdge==startEdge) // if we have seen all edges yet
        currentEdge=null;
        
        return  anObject;
    }
    }
    
    class AllEdgesEnumerator implements Enumeration 
    {
    private Edge  currentEdge;
    private Edge  startEdge;
    
    AllEdgesEnumerator(Edge start) {
            startEdge = start ;
        currentEdge=start;
        }
    
    public boolean  hasMoreElements() {
            return currentEdge != null;
        }
    
    public Object   nextElement() {
        Object  anObject = currentEdge;
        currentEdge = currentEdge.succ;
        if(currentEdge == startEdge)
        currentEdge = null;
        return  anObject;
    }
    }

    class AllNodesEnumerator implements Enumeration 
    {
    private Vertex  currentVertex;
    private Vertex  startVertex;
    
    AllNodesEnumerator(Vertex start) {
            startVertex = start;
        currentVertex = start ;
        }
    
    public boolean  hasMoreElements() {
            return currentVertex != null;
        }
    
    public Object   nextElement() {
        Object  anObject = currentVertex;
        currentVertex = currentVertex.succ;
        
        if(currentVertex==startVertex) // if we have seen all edges yet
        currentVertex=null;
        
        return  anObject;
    }
    }
    
    // start of Graph definition
    
    private Vertex m_Vertices;
    private Edge m_Edges;
    private int numOfVertices;
    private int numOfEdges;
    private Hashtable vhash;
    private VertexMap dfsnum;
    private VertexMap compnum;
    private boolean SCC_found;
    private int numOfSCCs;
    private Vector largestSCC;
    
    
    /** The constructor takes no arguments.
     */
    public Graph() 
    {
        numOfSCCs=0;
	    m_Vertices = null;
        m_Edges = null;
        numOfVertices=0;
        numOfEdges=0;
	    dfsnum = new VertexMap(this);
	    compnum = new VertexMap(this);
	    vhash = new Hashtable();
	    SCC_found = false;
	    largestSCC = new Vector();
    }
    
    /** Adds a new Vertex to the graph.
     * @param obj The object to associate with the Vertex
     * @return the newly created Vertex
     */
    public Vertex newVertex(Object obj) {
	SCC_found = false;
        Vertex v=new Vertex(obj);
        numOfVertices++;
        if(m_Vertices != null) { // if v is not the first Vertex to be inserted
        Vertex last=m_Vertices.pred;
        v.succ=m_Vertices;
        v.pred=last;
        m_Vertices.pred=v;
        last.succ=v;
        }
        m_Vertices=v; // m_Vertices points to the Vertex that was inserted last
        vhash.put(obj, v);
        return v;
    }
    
    /** Get a Vertex object by the object it holds.
     * @param obj The object to associate with the Vertex
     * @return the Vertex or null, if no vertex holds it
     */
    public Vertex getVertex(Object obj) {
        return (Vertex)vhash.get(obj);
    }
    
    /** Adds a new Edge from Vertex s to Vertex t.
        @param obj The object to associate with the Edge
        @param s source Vertex
        @param t target Vertex
        @return The newly created Edge
    */
    public Edge newEdge(Vertex s, Vertex t) {
	SCC_found = false;
        Edge e = new Edge();
        numOfEdges++;
        e.source=s;
        e.target=t;
        if(m_Edges!=null) { // if e is not the first Edge to be inserted
            Edge first=m_Edges.succ; // "first" Edge of the cyclic list of m_Edges
            Edge last=first.pred;
            e.succ=first;
            e.pred=last;
            first.pred=e;
            last.succ=e;
        }
        m_Edges=e; // m_Edges points to the last inserted Edge
        // insert e in the out-list of s
        if(s.out == null) 
            s.out=e;
        else {
            Edge first=s.out; // "first" Edge of the cyclic list of out-Edges
            Edge last=first.outpred;
            e.outsucc=first;
            e.outpred=last;
            first.outpred=e;
            last.outsucc=e;
        }
        // insert e in the in-list of t
        if(t.in == null) 
            t.in=e;
        else {
            Edge first=t.in; // "first" Edge of the cyclic list of in-Edges
            Edge last=first.inpred;
            e.insucc=first;
            e.inpred=last;
            first.inpred=e;
            last.insucc=e;
        }
        return e;
    }
    
    
    /** Returns the source Vertex of an Edge.
    @param  e  the Edge
    @return the source Vertex of e
    */
    public Vertex source(Edge e) {
        return e.source;
    }
    
    /** Returns the target Vertex of an Edge.
    @param  e  the Edge
    @return the target Vertex of e
    */
    public Vertex target(Edge e) {
        return e.target;
    }
    
    /** Returns the first Vertex of v
     */
    public Vertex firstVertex() {
        return m_Vertices;
    }
    
    /** Returns an enumeration of all Vertices which are adjacent to v.
     */
    public AdjVerticesEnumerator adjVertices(Vertex v) {
        return new AdjVerticesEnumerator(v);
    }
    
    /** Returns an enumeration of all Edges with source v.
     */
    public AdjEdgesEnumerator adjEdges(Vertex v) {
        return new AdjEdgesEnumerator(v);
    }
    
    /** Returns an enumeration of all Edges with source v.
     */
    public AdjEdgesEnumerator outEdges(Vertex v) {
        return new AdjEdgesEnumerator(v);
    }
    
    /** Returns an enumeration of all Edges with target v.
     */
    public InEdgesEnumerator inEdges(Vertex v) {
        return new InEdgesEnumerator(v);
    }
    
    /** Returns an enumeration of all Edges of the graph.
     */
    public AllEdgesEnumerator allEdges() {
        return new AllEdgesEnumerator(m_Edges);
    }
    
    /** Returns an enumeration of all Vertices of the graph.
     */
    public AllNodesEnumerator  allVertices() {
        return new AllNodesEnumerator(m_Vertices);
    }
        
    /** Returns the current number of Vertices of the graph.
     */
    public int numOfVertices() {
        return numOfVertices;
    }
    
    /** Returns the current number of Edges of the graph.
     */
    public int numOfEdges() {
        return numOfEdges;
    }
    
    /** Makes the graph the empty graph.
     */
    public void clear() {
        m_Vertices=null;
        m_Edges=null;
    }

    // SCC algorithm and related structs
    public class VertexMap extends Hashtable {
    
        Hashtable H;
        
        /**
        Introduces a variable M of type VertexMap and initializes
        it with a mapping m from the set of all nodes of G into
        the set of all objects.
        @param G The underlying graph.
        */ 
        public VertexMap(Graph G) {
            int n=G.numOfVertices();
            if(n == 0)
            n=10;
            H = new Hashtable(n);
        }
        
        /**
        Introduces a variable M of type VertexMap and initializes
        it with a mapping m from the set of all nodes of G into
        the set of int's. The variables in the range
        of m are initialized to value.
        @param G The underlying graph.
        @param value The value to associate with every node.
        */
        public VertexMap(Graph G, int value) {
            int n=G.numOfVertices();
            if(n == 0)
            n=10;
            H = new Hashtable(n);
            
            Vertex v;
            for(Enumeration enum=G.allVertices();enum.hasMoreElements();) {
                Integer newInt = new Integer(value);
                v=(Vertex) enum.nextElement();
                H.put(v,newInt);
            }
        }
        
        
        /**
        Sets the value of v in the mapping to obj.
        @param v The node.
        @param obj The object.
        */
        public void put(Vertex v,Object obj) {
            H.put(v,obj);
        }
        
        /**
        Returns the value of v in the mapping.
        @return The object associated with v.
        */
        public Object get(Vertex v) {
            return H.get(v);
        }
    
    }
    private int dfscount,compcount;
    
    public void sccDfs(Vertex v, Stack unfinished, Stack roots) {
        
        dfsnum.put(v,new Integer(dfscount++));
        Vertex w;
        Vector tmpVec = new Vector();
        
        // make v a tentative component of its own
        unfinished.push(v);
        roots.push(v);
        
        // explore all edges (v,w)
        for(Enumeration enum=adjVertices(v);enum.hasMoreElements();) {
            w=(Vertex) enum.nextElement();
            if(((Integer)dfsnum.get(w)).intValue() == -1)
            sccDfs(w,unfinished,roots);
            else
            if( ((Integer)compnum.get(w)).intValue() == -1)
                // merge components
                while( ((Integer)dfsnum.get((Vertex)roots.peek())).intValue() >
                ((Integer)dfsnum.get(w)).intValue())
                roots.pop();
        }
        
        // return from call for node v
        if( v == (Vertex) roots.peek()) {
            tmpVec.removeAllElements();
            do {
                w = (Vertex) unfinished.pop();
                compnum.put(w, new Integer(compcount));
                tmpVec.addElement(w.element());
            } while ( w != v);
            if (tmpVec.size() > largestSCC.size()) {
	        largestSCC = tmpVec;
	        //System.out.println("Largest SCC so far has " + largestSCC.size() + " vertices.");
	    }
            compcount++;
            roots.pop();
        }
    }
    
    public int findSCCs() {
        Stack unfinished = new Stack();
        Stack roots = new Stack();
        Vertex v;
        for(Enumeration enum=allVertices();enum.hasMoreElements();) {
            v = (Vertex) enum.nextElement();
            compnum.put(v, new Integer(-1));
            dfsnum.put(v, new Integer(-1));
        }
        dfscount=0;
        compcount=0;
        for(Enumeration enum=allVertices();enum.hasMoreElements();) {
            v = (Vertex) enum.nextElement();
            if( ((Integer)dfsnum.get(v)).intValue() == -1)
                sccDfs(v,unfinished,roots);
        }
        SCC_found = true;
        numOfSCCs=compcount;
        return compcount;
    }    

    public Vector getLargestSCC() {
	    return largestSCC;
    }
    
    public String toString() {
	if (!SCC_found)	    {
	    System.out.println("\n\n-I-     Start finding SCCs\n");
	    findSCCs();
	    System.out.println("\n\n-I-     End finding SCCs\n");
	}
        StringBuffer text = new StringBuffer();
	    text.append("The graph has " + numOfSCCs + " SCCs\n");
        for(Enumeration e=allVertices();e.hasMoreElements();) {
            Vertex v=(Vertex) e.nextElement();
            text.append("\n** Vertex: " + v.element().toString());
            text.append("  Outgoing Edges: ");
	    for (Enumeration f=outEdges(v);f.hasMoreElements();) {
		Edge ed=(Edge)f.nextElement();
		text.append(ed.toString()+ "  ");
	    }
            text.append("  Incoming Edges: ");
	    for (Enumeration f=inEdges(v);f.hasMoreElements();) {
		Edge ed=(Edge)f.nextElement();
		text.append(ed.toString()+ "  ");
	    }
	    text.append("  - Belongs to SCC " + compnum.get(v)  +"\n");
        }
       text.append("\nThe largest SCC contains " + largestSCC.size() + " Vertices:  ");
       for (int i=0; i<largestSCC.size();i++)
        text.append(largestSCC.elementAt(i).toString() + " ");
       text.append("\n");
        return text.toString();
    }

    /* method for testing this class */     
    public static void main(String[] ops) {
	Vector ver = new Vector();
	String tmpObj;
	System.out.println("\n\n-I-     Start testing Graph class\n");
	Graph g = new Graph();
	System.out.println("-I-     Creating vertices\n");
	for (int i=0; i<100; ++i) {
	    tmpObj = new String(String.valueOf(i));
	    ver.addElement(g.newVertex(tmpObj));
	}
	System.out.println("-I-     Creating edges\n");

	for (int j=3; j<60; j+=2) {
	    for (int k=1; k<30; k+=2) {
		g.newEdge((Vertex)ver.elementAt(j), (Vertex)ver.elementAt(k));
		    g.newEdge((Vertex)ver.elementAt(j+k), (Vertex)ver.elementAt(k));
	    }
	}

	for (int j=8; j<10; j++) {
	    for (int k=1; k<30; k+=2) {
		    g.newEdge((Vertex)ver.elementAt(k), (Vertex)ver.elementAt(j));
		    g.newEdge((Vertex)ver.elementAt(j+k), (Vertex)ver.elementAt(k));
	    }
	}

	for (int j=3; j<12; j++) {
	    for (int k=8; k<30; k+=1) {
		    g.newEdge((Vertex)ver.elementAt(k), (Vertex)ver.elementAt(j));
	    }
	}

	for (int j=0; j<7; j+=2) {
	    for (int k=10; k<31; k++) {
		    g.newEdge((Vertex)ver.elementAt(j), (Vertex)ver.elementAt(k));
		    g.newEdge((Vertex)ver.elementAt(j+k), (Vertex)ver.elementAt(k));
	    }
	}

	System.out.println("\n-I-     Printing graph\n");
	    
	System.out.println(g.toString());

    }

}    // end of Graph class
